#include <JGL/JGL.h>
#include <JGL/logger/logger.h>
#include <J3ML/Algorithm/Bezier.hpp>
#include "../internals/include/internals.h"

void JGL::J2D::Begin(RenderTarget* rt, bool clear_buffers) {
    GLfloat old_clear_color[4];
    if (rt != nullptr) {
        OpenGLState::render_target = rt;
        glGetFloatv(GL_COLOR_CLEAR_VALUE, old_clear_color);
        glGetIntegerv(GL_VIEWPORT, OpenGLState::viewport);
        OpenGLState::current_fbo = JGL::RenderTarget::GetActiveGLFramebufferHandle();
        JGL::RenderTarget::SetActiveGLRenderTarget(*rt);

        if (!(rt->GetJGLTexture()->GetFlags() & INVERT_Y))
            Logger::Warning("You're rendering onto a texture that is upside-down. Your draw commands won't work how you'd expect.");
    }
    glMatrixMode(GL_PROJECTION);
    glPushMatrix();
    glLoadIdentity();

    if (rt == nullptr)
        glViewport(0, 0, (int) OpenGLState::window_size.x, (int) OpenGLState::window_size.y),
        glOrtho(0, OpenGLState::window_size.x, OpenGLState::window_size.y, 0, -1, 1);
    else
        glOrtho(0, rt->GetDimensions().x, rt->GetDimensions().y, 0, -1, 1);

    glMatrixMode(GL_MODELVIEW);
    glPushMatrix();
    glLoadIdentity();

    //Get what the draw color was before we did anything.
    glGetFloatv(GL_CURRENT_COLOR, OpenGLState::oldColor);
    glColor4fv(OpenGLState::baseColor);

    glGetIntegerv(GL_ACTIVE_TEXTURE, &OpenGLState::activeTextureUnit);
    OpenGLState::activeTextureUnit = OpenGLState::activeTextureUnit - GL_TEXTURE0;

    if (OpenGLState::activeTextureUnit != 0) {
        glActiveTexture(GL_TEXTURE0);
        glClientActiveTexture(GL_TEXTURE0);
    }

    if (glIsEnabled(GL_DEPTH_TEST))
        OpenGLState::wasDepthTestEnabled = true,
        glDisable(GL_DEPTH_TEST);
    else
        OpenGLState::wasDepthTestEnabled = false;

    if (!glIsEnabled(GL_VERTEX_ARRAY))
        OpenGLState::wasVertexArraysEnabled = false,
        glEnableClientState(GL_VERTEX_ARRAY);
    else
        OpenGLState::wasVertexArraysEnabled = true;

    if (glIsEnabled(GL_NORMAL_ARRAY))
        OpenGLState::wasNormalArrayEnabled = true,
        glDisableClientState(GL_NORMAL_ARRAY);
    else
        OpenGLState::wasNormalArrayEnabled = false;

    if (!glIsEnabled(GL_CULL_FACE))
        OpenGLState::wasCullFaceEnabled = false,
        glEnable(GL_CULL_FACE),
        glCullFace(GL_BACK);
    else
        OpenGLState::wasCullFaceEnabled = true;

    if (!glIsEnabled(GL_BLEND))
        OpenGLState::wasBlendEnabled = false,
        glEnable(GL_BLEND),
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    else
        OpenGLState::wasBlendEnabled = true;

    if (glIsEnabled(GL_TEXTURE_2D))
        OpenGLState::wasTexture2DEnabled = true,
        glDisable(GL_TEXTURE_2D);

    if (glIsEnabled(GL_TEXTURE_COORD_ARRAY))
        OpenGLState::wasTextureCoordArrayEnabled = true;
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);

    if (glIsEnabled(GL_COLOR_ARRAY))
        OpenGLState::wasColorArrayEnabled = true,
        glDisableClientState(GL_COLOR_ARRAY);
    else
        OpenGLState::wasColorArrayEnabled = false;

    if (!inJ3D)
        inJ2D = true;
    else { Logger::Error("Beginning J2D context inside of J3D context?"); }

    if (rt != nullptr && clear_buffers) {
        glClearColor(rt->GetClearColor().RedChannelNormalized(), rt->GetClearColor().GreenChannelNormalized(), rt->GetClearColor().BlueChannelNormalized(), rt->GetClearColor().AlphaChannelNormalized());
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glClearColor(old_clear_color[0], old_clear_color[1], old_clear_color[2], old_clear_color[3]);
    }
}

void JGL::J2D::End() {
    //Change back to the previous projection (The 3D one in Re3D's case.)
    glPopMatrix();
    glMatrixMode(GL_PROJECTION);
    glPopMatrix();
    glMatrixMode(GL_MODELVIEW);
    if (OpenGLState::wasDepthTestEnabled)
        glEnable(GL_DEPTH_TEST);

    if (!OpenGLState::wasVertexArraysEnabled)
        glDisableClientState(GL_VERTEX_ARRAY);

    if (!OpenGLState::wasCullFaceEnabled)
        glDisable(GL_CULL_FACE);

    if (!OpenGLState::wasBlendEnabled)
        glDisable(GL_BLEND);

    // Select whatever texture mapper was selected before.
    glActiveTexture(GL_TEXTURE0 + OpenGLState::activeTextureUnit);
    glClientActiveTexture(GL_TEXTURE0 + OpenGLState::activeTextureUnit);

    if (OpenGLState::wasTexture2DEnabled)
        glEnable(GL_TEXTURE_2D);

    if (OpenGLState::wasTextureCoordArrayEnabled)
        glEnableClientState(GL_TEXTURE_COORD_ARRAY);

    if (OpenGLState::wasNormalArrayEnabled)
        glEnableClientState(GL_NORMAL_ARRAY);

    if (OpenGLState::wasColorArrayEnabled)
        glEnableClientState(GL_COLOR_ARRAY);

    //Put the draw color back how it was before.
    glColor4fv(OpenGLState::oldColor);

    if (OpenGLState::render_target != nullptr) {
        OpenGLState::render_target->MSAABlit();
        OpenGLState::render_target = nullptr;
        glBindFramebuffer(GL_FRAMEBUFFER, OpenGLState::current_fbo);
        glViewport(OpenGLState::viewport[0], OpenGLState::viewport[1], OpenGLState::viewport[2], OpenGLState::viewport[3]);
    }
    inJ2D = false;
}

void JGL::J2D::RequiredLight(const JGL::LightBase* light) {
    bool success = false;
    for (auto& item : required_lights)
        if (item == nullptr) {
            item = light;
            success = true;
            break;
        }

    if (!success)
        Logger::Error("You cannot specify more than 8 required lights.");
}

void JGL::J2D::OptionalLights(const LightBase** lights, const size_t& light_count) {
    for (size_t i = 0; i < light_count; i++)
        light_array.push_back(lights[i]);
}

void JGL::J2D::DrawPoint(const Color4& color, const Vector2& coordinates, float radius) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    glPointSize(radius);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), coordinates.ptr());
    glDrawArrays(GL_POINTS, 0, 1);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::DrawPoint(const Color4& color, float x, float y, float radius) {
    DrawPoint(color, {x, y}, radius);
}

void JGL::J2D::DrawLine(const Color4& color, const Vector2& A, const Vector2& B, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");
    Vector2 vertices[] = {A, B};

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_LINES, 0, 2);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::DrawLine(const Color4& color, float x, float y, float w, float h, float thickness) {
    JGL::J2D::DrawLine(color, {x, y}, {w, h}, thickness);
}

void JGL::J2D::DrawDottedLine(const Color4& color, const Vector2& A, const Vector2& B, float spacing, float thickness) {
    float distance = Vector2::Distance(A, B);
    Vector2 direction = (B - A).Normalized();

    unsigned int point_count = distance / spacing;
    std::vector<Vector2> points(point_count);

    if (spacing > distance)
        Logger::Error("Drawing a dotted line that would have no dots?");

    for (unsigned int i = 0; i < point_count; ++i)
        points[i] = A + direction * (i * spacing);

    return JGL::J2D::DrawPoints(color, points.data(), points.size(), thickness);
}

void JGL::J2D::DrawDottedLine(const Color4& color, float x1, float y1, float x2, float y2, float spacing, float thickness) {
    return JGL::J2D::DrawDottedLine(color, {x1, y1}, {x2, y2}, spacing, thickness);
}

void JGL::J2D::DrawDashedLine(const Color4& color, const Vector2& A, const Vector2& B, float spacing, float dash_length, float thickness) {
    float distance = Vector2::Distance(A, B);
    Vector2 direction = (B - A).Normalized();
    float length_of_dash_and_gap = dash_length + spacing;
    unsigned int dash_count = distance / length_of_dash_and_gap;

    if (spacing > distance)
        Logger::Error("Drawing a dashed line that would have no dashes?");

    Vector2 A_current, B_current;
    for (unsigned int i = 0; i < dash_count; i++) {
        A_current = A + direction * (i * length_of_dash_and_gap);
        B_current = A_current + (direction * dash_length);
        JGL::J2D::DrawLine(color, A_current, B_current, thickness);
    }

    // For the little piece at the end.
    float distance_left = distance - (dash_count * length_of_dash_and_gap);
    if (distance_left > 0) {
        A_current = A + direction * (dash_count * length_of_dash_and_gap);
        B_current = A_current + direction * std::min(dash_length, distance_left);
        JGL::J2D::DrawLine(color, A_current, B_current, thickness);
    }
}

void JGL::J2D::DrawDashedLine(const Color4& color, float x1, float y1, float x2, float y2, float spacing, float dash_length, float thickness) {
    return JGL::J2D::DrawDashedLine(color, {x1, y1}, {x2, y2}, spacing, dash_length, thickness);
}

void JGL::J2D::DrawGradientLine(const Color4& color1, const Color4& color2, const Vector2& A, const Vector2& B, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {A, B};
    GLfloat colors[8] = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                         color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized() };

    glEnableClientState(GL_COLOR_ARRAY);
    glLineWidth(thickness);
    glColorPointer(4,GL_FLOAT,sizeof(Color4), colors);
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_LINES, 0, 2);
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4fv(OpenGLState::baseColor);
}

void DrawGradientLine(const Color4& color1, const Color4& color2, float x, float y, float w, float h, float thickness) {
    JGL::J2D::DrawGradientLine(color1, color2, {x, y}, {w, h}, thickness);
}

void JGL::J2D::OutlineRect(const Color4& color, const Vector2& pos, const Vector2& size, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_LINE_LOOP, 0, 4);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillRect(const Color4& color, const Vector2& pos, const Vector2& size) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_QUADS, 0, 4);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillGradientRect(const Color4& color1, const Color4& color2, const Direction& gradient, const Vector2& pos, const Vector2& size) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{pos.x, pos.y}, {pos.x, pos.y + size.y}, {pos.x + size.x, pos.y + size.y}, {pos.x + size.x, pos.y}};
    std::vector<GLfloat> colors{};

    if (gradient == Direction::Horizontal)
        colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),
                  color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

    else if (gradient == Direction::Vertical)
        colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                  color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(), color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized()};

    else if (gradient == Direction::Diagonal_SWNE)
        colors = {(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f,
                  color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(), (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                  (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized()};

    else if (gradient == Direction::Diagonal_NWSE)
        colors = {color1.RedChannelNormalized(), color1.GreenChannelNormalized(), color1.BlueChannelNormalized(), color1.AlphaChannelNormalized(),(color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f,
                  (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f, (color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f, color2.RedChannelNormalized(), color2.GreenChannelNormalized(), color2.BlueChannelNormalized(), color2.AlphaChannelNormalized(),
                  (color1.RedChannelNormalized() + color2.RedChannelNormalized()) / 2.f, (color1.GreenChannelNormalized() + color2.GreenChannelNormalized()) / 2.f, (color1.BlueChannelNormalized() + color2.BlueChannelNormalized()) / 2.f,(color1.AlphaChannelNormalized() + color2.AlphaChannelNormalized()) / 2.f};

    glEnableClientState(GL_COLOR_ARRAY);
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glColorPointer(4, GL_FLOAT, 0, colors.data());
    glDrawArrays(GL_QUADS, 0, 4);
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius, unsigned int subdivisions) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    JGL::J2D::FillRect(color, {pos.x + radius, pos.y}, {size.x - 2 * radius, size.y});
    JGL::J2D::FillRect(color, {pos.x, pos.y + radius}, {size.x, size.y - 2 * radius});

    JGL::J2D::FillCircle(color, {pos.x + radius, pos.y + radius}, radius, subdivisions);
    JGL::J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + radius}, radius, subdivisions);
    JGL::J2D::FillCircle(color, {pos.x + radius, pos.y + size.y - radius}, radius, subdivisions);
    JGL::J2D::FillCircle(color, {pos.x + size.x - radius, pos.y + size.y - radius}, radius, subdivisions);
}

void JGL::J2D::DrawSprite(const Texture* texture, float positionX, float positionY, float rad_rotation,
                     float originX, float originY,float scaleX, float scaleY,
                     const Color4& color, Direction inversion) {
    DrawSprite(*texture, {positionX, positionY}, rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
}
void JGL::J2D::DrawSprite(const Texture* texture, const Vector2& position, float rad_rotation, const Vector2& origin,
                     const Vector2& scale, const Color4& color, Direction inversion) {

    DrawSprite(*texture, position, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawPartialSprite(const Texture* texture, const Vector2& position, const Vector2& sub_texture_position,
                            const Vector2& sub_texture_size, float rad_rotation, const Vector2& origin,
                            const Vector2& scale, const Color4& color, Direction inversion) {
    DrawPartialSprite(*texture, position, sub_texture_position, sub_texture_size, rad_rotation, origin, scale, color, inversion);

}

void JGL::J2D::DrawPartialSprite(const Texture* texture, float positionX, float positionY, float sub_texture_positionX,
                            float sub_texture_positionY, unsigned int sub_texture_sizeX,
                            unsigned int sub_texture_sizeY, float rad_rotation, float originX, float originY,
                            float scaleX, float scaleY, const Color4& color, Direction inversion) {
    DrawPartialSprite(*texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY}, {(float) sub_texture_sizeX, (float) sub_texture_sizeY},
                      rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
}

void JGL::J2D::DrawMirrorSprite(const Texture* texture, const Vector2& position, Direction mirror_axis, float rad_rotation,
                           const Vector2& origin, const Vector2& scale, const Color4& color) {
    DrawMirrorSprite(*texture, position, mirror_axis, rad_rotation, origin, scale, color);
}

void JGL::J2D::DrawSprite(const RenderTarget* render_target, const Vector2& position, float rad_rotation,
                     const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
    DrawSprite(*render_target, position, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawRenderTarget(const RenderTarget* render_target, const Vector2& position, float rad_rotation,
                           const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
    DrawSprite(*render_target, position, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawRenderTarget(const RenderTarget& render_target, const Vector2& position, float rad_rotation,
                           const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
    DrawSprite(render_target, position, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawSprite(const JGL::RenderTarget& rt, const Vector2& position, float rad_rotation, const Vector2& origin,
                     const Vector2& scale, const Color4& color, Direction inversion) {

    //Correct for the render-target being upside-down.
    Direction d{};
    if (inversion == Direction::None && !(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
        d = Direction::Vertical;
    else if (inversion == Direction::Horizontal) {
        d = Direction::Horizontal;
        if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
            d = Direction::Horizontal | Direction::Vertical;
    }
    else if (inversion& Direction::Horizontal && inversion& Direction::Vertical) {
        d = Direction::Horizontal;
        if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
            d = Direction::Horizontal | Direction::Vertical;
    }

    //Change the blending mode such that the alpha doesn't get multiplied again.
    if (rt.OwnsTexture())
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    JGL::J2D::DrawPartialSprite(*rt.GetJGLTexture(), position, {0, 0}, Vector2(rt.GetDimensions()), rad_rotation, origin, scale, color, d);

    if (rt.OwnsTexture())
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void JGL::J2D::DrawSprite(const Texture& texture, const Texture &alpha_mask, const Vector2& position, float rad_rotation,
                     const Vector2& origin, const Vector2& scale,const Color4& color, JGL::Direction inversion) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    if (texture.GetDimensions() != alpha_mask.GetDimensions())
        Logger::Warning("The alpha mask and the texture are not the same size.");

    const Vector2 size = Vector2(texture.GetDimensions());

    std::array<Vector2, 4> textureCoordinates{};
    if (texture.GetFlags() & INVERT_Y)
        textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(1, 0), Vector2(1, 1)};
    else
        textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

    // TODO: Kind of a mess, refactor to be more sensible later.
    // Factors in scaling and origin correctly.
    // i.e. to render at 2x size, from the center, at coords XY, use {2, 2} scale, and {0.5, 0.5} offset.
    const Vector2 offset = origin * size;
    Vector2 pos2 = position;
    Vector2 scaled_size = scale * size;
    Vector2 size2 = scaled_size;
    float cos_theta = std::cos(rad_rotation);
    float sin_theta = std::sin(rad_rotation);

    std::array<Vector2, 4> vertices =
            {
                    pos2,                                 // Top-left vertex
                    {pos2.x, pos2.y + size2.y},            // Bottom-left
                    {pos2.x + size2.x, pos2.y + size2.y},   // Bottom-right
                    {pos2.x + size2.x, pos2.y}             // Top-right
            };

    //Rotate the vertices about the origin by float rad_rotation.
    if (rad_rotation != 0)
        for (auto& v: vertices)
            v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                 pos2.x + offset.x * scale.x,
                 (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                 pos2.y + offset.y * scale.y
            };

    if (inversion == Direction::Vertical)
        std::swap(textureCoordinates[0], textureCoordinates[1]),
                std::swap(textureCoordinates[3], textureCoordinates[2]);
    if (inversion == Direction::Horizontal)
        std::swap(textureCoordinates[0], textureCoordinates[3]),
                std::swap(textureCoordinates[1], textureCoordinates[2]);


    glColor4ubv(color.ptr());

    // Texture 0.
    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());

    // Texture 1.
    glActiveTexture(GL_TEXTURE1);
    glClientActiveTexture(GL_TEXTURE1);
    glEnable(GL_TEXTURE_2D);
    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnableClientState(GL_VERTEX_ARRAY);
    glTexEnvi(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

    glBindTexture(GL_TEXTURE_2D, alpha_mask.GetGLTextureHandle());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());

    // Draw.
    glDrawArrays(GL_QUADS, 0, 4);

    // Reset Texture 1.
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisableClientState(GL_VERTEX_ARRAY);
    glDisable(GL_TEXTURE_2D);

    // Reset Texture 0.
    glActiveTexture(GL_TEXTURE0);
    glClientActiveTexture(GL_TEXTURE0);
    glBindTexture(GL_TEXTURE_2D, 0);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisable(GL_TEXTURE_2D);

    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::DrawSprite(const Texture* texture, const Texture* alpha_mask, const Vector2& position, float rad_rotation,
                     const Vector2& origin, const Vector2& scale,const Color4& color, Direction inversion) {
    DrawSprite(*texture, *alpha_mask, position, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawSprite(const Texture& texture, const Texture& alpha_mask, float positionX, float positionY, float rad_rotation,
                     float originX, float originY,float scaleX, float scaleY,const Color4& color, Direction inversion) {
    DrawSprite(texture, alpha_mask, {positionX, positionY}, rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
}

void JGL::J2D::DrawSprite(const Texture* texture, const Texture* alpha_mask, float positionX, float positionY, float rad_rotation,
                     float originX, float originY,float scaleX, float scaleY,const Color4& color, Direction inversion) {
    DrawSprite(*texture, *alpha_mask, {positionX, positionY}, rad_rotation, {originX, originY}, {scaleX, scaleY}, color, inversion);
}

void JGL::J2D::DrawPartialRenderTarget(const JGL::RenderTarget& rt, const Vector2& position, const Vector2& sub_texture_position, const Vector2& sub_texture_size,
                                  float rad_rotation, const Vector2& origin, const Vector2& scale, const Color4& color, JGL::Direction inversion) {

    //Correct for the render-target being upside-down.
    Direction d{};
    if (inversion == Direction::None && !(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
        d = Direction::Vertical;

    else if (inversion == Direction::Horizontal) {
        d = Direction::Horizontal;
        if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
            d = Direction::Horizontal | Direction::Vertical;
    }
    else if (inversion& Direction::Horizontal && inversion& Direction::Vertical) {
        d = Direction::Horizontal;
        if (!(rt.GetJGLTexture()->GetFlags() & INVERT_Y))
            d = Direction::Horizontal | Direction::Vertical;
    }

    //Change the blending mode such that the alpha doesn't get multiplied again.
    if (rt.OwnsTexture())
        glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_ALPHA);

    JGL::J2D::DrawPartialSprite(*rt.GetJGLTexture(), position, sub_texture_position, sub_texture_size, rad_rotation, origin, scale, color, d);

    if (rt.OwnsTexture())
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
}

void JGL::J2D::DrawPartialRenderTarget(const JGL::RenderTarget* rt, const Vector2& position,const Vector2& sub_texture_position,const Vector2& sub_texture_size, float rad_rotation,
                                  const Vector2& origin, const Vector2& scale, const Color4& color, Direction inversion) {
    DrawPartialRenderTarget(*rt, position, sub_texture_position, sub_texture_size, rad_rotation, origin, scale, color, inversion);
}

void JGL::J2D::DrawSprite(const Texture& texture, const Vector2& pos, float rad_rotation, const Vector2& origin,
                     const Vector2& scale, const Color4& color, Direction inversion) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");


    const Vector2 size = Vector2(texture.GetDimensions());

    std::array<Vector2, 4> textureCoordinates{};
    if (texture.GetFlags() & INVERT_Y)
        textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(1, 0), Vector2(1, 1)};
    else
        textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};

    // TODO: Kind of a mess, refactor to be more sensible later.
    // Factors in scaling and origin correctly.
    // i.e. to render at 2x size, from the center, at coords XY, use {2, 2} scale, and {0.5, 0.5} offset.
    const Vector2 offset = origin * size;
    Vector2 pos2 = pos;
    Vector2 scaled_size = scale * size;
    Vector2 size2 = scaled_size;
    float cos_theta = std::cos(rad_rotation);
    float sin_theta = std::sin(rad_rotation);

    std::array<Vector2, 4> vertices =
            {
                    pos2,                                 // Top-left vertex
                    {pos2.x, pos2.y + size2.y},            // Bottom-left
                    {pos2.x + size2.x, pos2.y + size2.y},   // Bottom-right
                    {pos2.x + size2.x, pos2.y}             // Top-right
            };

    //Rotate the vertices about the origin by float rad_rotation.
    if (rad_rotation != 0)
        for (auto& v: vertices)
            v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                 pos2.x + offset.x * scale.x,
                 (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                 pos2.y + offset.y * scale.y
            };

    if (inversion == Direction::Vertical)
        std::swap(textureCoordinates[0], textureCoordinates[1]),
                std::swap(textureCoordinates[3], textureCoordinates[2]);
    if (inversion == Direction::Horizontal)
        std::swap(textureCoordinates[0], textureCoordinates[3]),
                std::swap(textureCoordinates[1], textureCoordinates[2]);

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_TEXTURE_2D);
    glColor4ubv(color.ptr());
    glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
    glDrawArrays(GL_QUADS, 0, 4);
    glBindTexture(GL_TEXTURE_2D, 0);
    glColor4fv(OpenGLState::baseColor);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisable(GL_TEXTURE_2D);
}


void JGL::J2D::DrawSprite(const Texture& texture, float positionX, float positionY, float rad_rotation, float originX,
                     float originY, float scaleX, float scaleY, const Color4& color, Direction inversion) {
    DrawSprite(texture,
               {positionX, positionY},
               rad_rotation,
               {originX, originY},
               {scaleX, scaleY},
               color, inversion);
}

void JGL::J2D::DrawPartialSprite(const Texture& texture, const Vector2& position, const Vector2& sub_texture_position,
                            const Vector2& sub_texture_size, float rad_rotation, const Vector2& origin,
                            const Vector2& scale, const Color4& color, Direction inversion) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    const Vector2 textureSize = Vector2(texture.GetDimensions());
    std::array<GLfloat, 8> textureCoordinates{};
    // Calculate texture coordinates (relative to the whole texture)
    if (!(texture.GetFlags() & INVERT_Y))
        textureCoordinates = {
                sub_texture_position.x / textureSize.x,
                sub_texture_position.y / textureSize.y,
                sub_texture_position.x / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                (sub_texture_position.y + sub_texture_size.y) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                sub_texture_position.y / textureSize.y
        };

    else {
        textureCoordinates = {
                sub_texture_position.x / textureSize.x,
                (textureSize.y - sub_texture_position.y) / textureSize.y,
                sub_texture_position.x / textureSize.x,
                (textureSize.y - (sub_texture_position.y + sub_texture_size.y)) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                (textureSize.y - (sub_texture_position.y + sub_texture_size.y)) / textureSize.y,
                (sub_texture_position.x + sub_texture_size.x) / textureSize.x,
                (textureSize.y - sub_texture_position.y) / textureSize.y
        };
    }

    if (inversion& Direction::Vertical)
        std::swap(textureCoordinates[1], textureCoordinates[3]),
                std::swap(textureCoordinates[5], textureCoordinates[7]);

    if (inversion& Direction::Horizontal)
        std::swap(textureCoordinates[0], textureCoordinates[6]),
                std::swap(textureCoordinates[2], textureCoordinates[4]);

    const Vector2 offset = origin * sub_texture_size;
    Vector2 pos2 = position;
    Vector2 scaled_size = scale * sub_texture_size;
    Vector2 size2 = scaled_size;
    float cos_theta = std::cos(rad_rotation);
    float sin_theta = std::sin(rad_rotation);

    std::array<Vector2, 4> vertices =
            {
                    pos2,                                // Top-left
                    {pos2.x, pos2.y + size2.y},          // Bottom-left
                    {pos2.x + size2.x, pos2.y + size2.y},// Bottom-right
                    {pos2.x + size2.x, pos2.y}           // Top-right
            };

    //Rotate the vertices about the origin by float rad_rotation.
    if (rad_rotation != 0)
        for (auto& v: vertices)
            v = {(v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta +
                 pos2.x + offset.x * scale.x,
                 (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta +
                 pos2.y + offset.y * scale.y};

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_TEXTURE_2D);
    glColor4ubv(color.ptr());
    glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glTexCoordPointer(2, GL_FLOAT, 0, textureCoordinates.data());
    glDrawArrays(GL_QUADS, 0, 4);
    glBindTexture(GL_TEXTURE_2D, 0);
    glColor4fv(OpenGLState::baseColor);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisable(GL_TEXTURE_2D);
}

void JGL::J2D::DrawPartialSprite(const JGL::Texture& texture, float positionX, float positionY, float sub_texture_positionX,
                            float sub_texture_positionY, unsigned int sub_texture_sizeX,
                            unsigned int sub_texture_sizeY, float originX, float originY, float rad_rotation,
                            float scaleX, float scaleY, const Color4& color, JGL::Direction inversion) {

    JGL::J2D::DrawPartialSprite(texture, {positionX, positionY}, {sub_texture_positionX, sub_texture_positionY},
                           {(float) sub_texture_sizeX, (float) sub_texture_sizeY}, rad_rotation, {originX, originY},
                           {scaleX, scaleY}, color, inversion);
}

void JGL::J2D::DrawMirrorSprite(const Texture& texture, const Vector2& position, Direction mirror_axis, float rad_rotation, const Vector2& origin, const Vector2& scale, const Color4& color) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    if (mirror_axis == Direction::None)
        Logger::Warning("Drawing non-mirrored sprite with JGL::J2D::DrawMirrorSprite?");

    glBindTexture(GL_TEXTURE_2D, texture.GetGLTextureHandle());
    Vector2 size = Vector2(texture.GetDimensions());

    std::array<Vector2, 4> textureCoordinates  = {Vector2(0, 0), Vector2(0, 1), Vector2(1, 1), Vector2(1, 0)};
    if (mirror_axis == Direction::Horizontal) {
        size.x *= 2;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
        if (texture.GetFlags() & INVERT_Y)
            textureCoordinates = {Vector2(0, 1), Vector2(0, 0), Vector2(2, 0), Vector2(2, 1)};
        else
            textureCoordinates = {Vector2(0, 0), Vector2(0, 1), Vector2(2, 1), Vector2(2, 0)};
    }

    else if (mirror_axis == Direction::Vertical) {
        size.y *= 2;
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
        if (texture.GetFlags() & INVERT_Y)
            textureCoordinates = {Vector2(0, 2), Vector2(0, 0), Vector2(1, 0), Vector2(1, 2)};
        else
            textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(1, 2), Vector2(1, 0)};
    }

    else if ((mirror_axis& Direction::Horizontal) && (mirror_axis& Direction::Vertical)) {
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_MIRRORED_REPEAT);
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_MIRRORED_REPEAT);
        size *= 2;
        if (texture.GetFlags() & INVERT_Y)
            textureCoordinates = {Vector2(0, 2), Vector2(0, 0), Vector2(2, 0), Vector2(2, 2)};
        else
            textureCoordinates = {Vector2(0, 0), Vector2(0, 2), Vector2(2, 2), Vector2(2, 0)};
    }

    const Vector2 offset = origin * size;
    Vector2 pos2 = position;
    Vector2 size2 = scale * size;
    float cos_theta = std::cos(rad_rotation);
    float sin_theta = std::sin(rad_rotation);

    std::array<Vector2, 4> vertices =
            {
                    pos2,
                    {pos2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y + size2.y},
                    {pos2.x + size2.x, pos2.y}
            };

    if (rad_rotation != 0)
        for (auto& v : vertices)
            v = {
                    (v.x - pos2.x - offset.x * scale.x) * cos_theta - (v.y - pos2.y - offset.y * scale.y) * sin_theta + pos2.x + offset.x * scale.x,
                    (v.x - pos2.x - offset.x * scale.x) * sin_theta + (v.y - pos2.y - offset.y * scale.y) * cos_theta + pos2.y + offset.y * scale.y
            };

    glEnableClientState(GL_TEXTURE_COORD_ARRAY);
    glEnable(GL_TEXTURE_2D);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glTexCoordPointer(2, GL_FLOAT, sizeof(Vector2), textureCoordinates.data());
    glDrawArrays(GL_QUADS, 0, 4);

    //Reset the wrapping mode.
    if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_EDGE)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    else if (texture.GetWrappingMode() == TextureWrappingMode::REPEAT)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    else if (texture.GetWrappingMode() == TextureWrappingMode::CLAMP_TO_BORDER)
        glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_BORDER),
                glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_BORDER);
    glBindTexture(GL_TEXTURE_2D, 0);
    glColor4fv(OpenGLState::baseColor);
    glDisableClientState(GL_TEXTURE_COORD_ARRAY);
    glDisable(GL_TEXTURE_2D);
}

void JGL::J2D::OutlineCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    float step = (2.f * Math::Pi) / (float) subdivisions;
    std::vector<Vector2> vertices(subdivisions);
    GLfloat angle, x, y;

    int i = 0;
    for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step)
    {
        x = radius * std::sin(angle) + center.x;
        y = radius * std::cos(angle) + center.y;
        if (i < subdivisions)
            vertices[i] = {x, y};
        else
            vertices.emplace_back(x, y);
        i++;
    }

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glDrawArrays(GL_LINE_LOOP, 0, (int) vertices.size());
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillCircle(const Color4& color, const Vector2& center, float radius, unsigned int subdivisions) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    GLfloat angle, x, y;
    float step =  (2.f * Math::Pi) / (float) subdivisions;
    std::vector<Vector2> vertices(subdivisions);

    /* Most of the time the vector "vertices" is either the same size or size + 1 of the number of subdivisions.
     * Because a float is a decimal, It'd take too long to get rid of the emplace-back making us
     * wait around for the container to resize. This gets rid of it for what we can guarantee. */
    int i = 0;
    for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
        x = radius * std::sin(angle) + center.x;
        y = radius * std::cos(angle) + center.y;
        if (i < subdivisions)
            vertices[i] = {x, y};
        else
            vertices.emplace_back(x, y);
        i++;
    }

    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glDrawArrays(GL_TRIANGLE_FAN, 0, (int) vertices.size());
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::OutlineTriangle(const Color4& color, const Triangle2D& tri, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_LINE_LOOP, 0, 3);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillTriangle(const Color4& color, const Triangle2D& tri) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};

    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Triangle2D& tri) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    Vector2 vertices[] = {{tri.A.x, tri.A.y}, {tri.B.x, tri.B.y}, {tri.C.x, tri.C.y}};
    GLfloat colors[] = {a_color.RedChannelNormalized(), a_color.GreenChannelNormalized(), a_color.BlueChannelNormalized(), a_color.AlphaChannelNormalized(),
                        b_color.RedChannelNormalized(),b_color.GreenChannelNormalized(), b_color.BlueChannelNormalized(), b_color.AlphaChannelNormalized(),
                        c_color.RedChannelNormalized(), c_color.GreenChannelNormalized(), c_color.BlueChannelNormalized(),c_color.AlphaChannelNormalized()};

    glEnableClientState(GL_COLOR_ARRAY);
    glColorPointer(4, GL_FLOAT, sizeof(Color4), colors);
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_TRIANGLES, 0, 3);
    glDisableClientState(GL_COLOR_ARRAY);
    glColor4fv(OpenGLState::baseColor);
}

//TODO render all in once pass with GL_LINE_LOOP instead of separate lines.
void JGL::J2D::DrawCubicBezierCurve(const Color4& color, const Vector2& controlA, const Vector2& pointA, const Vector2& pointB, const Vector2& controlB,
                               int subdivisions, float thickness) {


    Vector2 last = controlA;
    const Vector2& first = controlB;
    for (int i = 0; i < subdivisions; ++i)
    {
        float alpha = (float) i / (float) subdivisions;
        Vector2 step = J3ML::Algorithm::Bezier(alpha, controlA, pointA, pointB, controlB);
        DrawLine(color, last, step, thickness);
        last = step;
    }

    // Have to manually draw the last segment of the curve.
    DrawLine(color, last, first, thickness);
}

void JGL::J2D::OutlinePolygon(const Color4& color, const Vector2* points, int points_size, float thickness) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    if (points[0] != points[points_size -1])
        throw std::runtime_error("JGL::J2D::OutlinePolygon: The first point and the last point must connect.");

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), points);
    glDrawArrays(GL_LINE_LOOP, 0, points_size);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::DrawGradientLine(const Color4& color_a, const Color4& color_b, float x, float y, float w, float h,
                           float thickness) {
    DrawGradientLine(color_a, color_b, {x, y}, {w, h}, thickness);
}

void JGL::J2D::DrawArc(const Color4& color, const Vector2& center, float radius, float arc_begin, float arc_end, unsigned int subdivisions, float thickness)
{
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    if (arc_begin == arc_end)
        Logger::Error("WTF are you even drawing???");

    if ( J3ML::Math::Mod(J3ML::Math::Abs(arc_begin-arc_end), J3ML::Math::Pi*2.f) == 0)
        Logger::Error("Use OutlineCircle instead!!");

    float step = Math::Abs(arc_end-arc_begin) / (float) subdivisions;
    std::vector<Vector2> vertices(subdivisions);
    GLfloat angle, x, y;

    int i = 0;
    for (angle = arc_begin; angle <= arc_end; angle += step) {
        x = center.x + (radius * std::cos(angle)); // TODO: Why is it fine here, but must be x,y flipped for sin,cos in Circle code?
        y = center.y + (radius * std::sin(angle));
        if (i < subdivisions)
            vertices[i] = {x, y};
        else
            vertices.emplace_back(x, y);
        i++;
    }

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glDrawArrays(GL_LINE_STRIP, 0, (int) vertices.size());
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::OutlineRoundedRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius, float thickness)
{
    // A rounded rectangle of size 2a x 2b with rounding radius r is given by
    // f(x; a, r) + f(y; b, r) = 1
    // where
    // f(x; a, r) = {
    // if |x| >= a - r: (|x| - (a-r) / r )^2
    // otherwise: 0
    // }

    //std::vector<Vector2> vertices;


    // TODO: Calculate vertices for top-left     quarter-circle.

    Vector2 tl = pos;
    Vector2 tr = pos + Vector2(size.x, 0);
    Vector2 br = pos + size;
    Vector2 bl = pos + Vector2(0, size.y);

    // Center-point around which to calculate each 'quarter-circle'.
    Vector2 anchor_tl = pos + Vector2(radius, radius);
    Vector2 anchor_tr = pos + Vector2(size.x-radius, radius);
    Vector2 anchor_br = pos + size - Vector2(radius, radius);
    Vector2 anchor_bl = pos + Vector2(radius, size.y - radius);

    //JGL::J2D::Begin();
    //JGL::J2D::DrawPoints(Colors::Red, {tl, tr, br, bl}, 2);
    //JGL::J2D::DrawPoints(Colors::Blue, {anchor_tl, anchor_tr, anchor_br, anchor_bl}, 2);
    //JGL::J2D::End();

    Vector2 anchor_topleft_top = pos + Vector2(radius, 0);
    Vector2 anchor_topright_top = pos + Vector2(size.x - radius, 0);
    // TODO: Calculate vertices for top-right    quarter-circle.
    Vector2 anchor_topright_right = pos + Vector2(size.x, radius);
    Vector2 anchor_bottomright_right = pos + Vector2(size.x, size.y - radius);
    // TODO: Calculate vertices for bottom-right quarter-circle.
    Vector2 anchor_bottomright_bottom = pos + Vector2(size.x - radius, size.y);
    Vector2 anchor_bottomleft_bottom = pos + Vector2(radius, size.y);
    // TODO: Calculate vertices for bottom-left  quarter-circle.
    Vector2 anchor_bottomleft_left = pos + Vector2(0, size.y - radius);
    Vector2 anchor_topleft_left = pos + Vector2(0, radius);

    //JGL::J2D::Begin();

    // The 3.01f, etc is a tiny-bit of overshoot to compensate for the fact that
    // this is not being plotted as a continuous line-loop.
    // Once i'm done working this out I expect bill will want to make it such.

    unsigned int subdivisions = 9;

    JGL::J2D::DrawArc(color, anchor_tl, radius, Math::Pi, 3.01f*Math::Pi/2.f, subdivisions, thickness);
    JGL::J2D::DrawLine(color, anchor_topleft_top, anchor_topright_top, thickness);
    JGL::J2D::DrawArc(color, anchor_tr, radius, 3.f*Math::Pi/2.f, 2.02*Math::Pi, subdivisions, thickness);
    JGL::J2D::DrawLine(color, anchor_topright_right, anchor_bottomright_right, thickness);
    JGL::J2D::DrawArc(color, anchor_br, radius, 0.0f, 1.01f*Math::Pi/2, subdivisions, thickness);
    JGL::J2D::DrawLine(color, anchor_bottomright_bottom, anchor_bottomleft_bottom, thickness);
    JGL::J2D::DrawArc(color, anchor_bl, radius, Math::Pi/2, Math::Pi*1.01f, subdivisions, thickness);
    JGL::J2D::DrawLine(color, anchor_bottomleft_left, anchor_topleft_left, thickness);
    //JGL::J2D::End();


}

void JGL::J2D::FillChamferRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius) {
    FillRoundedRect(color, pos, size, radius, 4);
}

void JGL::J2D::OutlineChamferRect(const Color4& color, const Vector2& pos, const Vector2& size, float radius,
                             float thickness) {
    Vector2 anchor_topleft_top = pos + Vector2(radius, 0);
    Vector2 anchor_topright_top = pos + Vector2(size.x - radius, 0);
    Vector2 anchor_topright_right = pos + Vector2(size.x, radius);
    Vector2 anchor_bottomright_right = pos + Vector2(size.x, size.y - radius);
    Vector2 anchor_bottomright_bottom = pos + Vector2(size.x - radius, size.y);
    Vector2 anchor_bottomleft_bottom = pos + Vector2(radius, size.y);
    Vector2 anchor_bottomleft_left = pos + Vector2(0, size.y - radius);
    Vector2 anchor_topleft_left = pos + Vector2(0, radius);

    Vector2 vertices[] = {anchor_topleft_top, anchor_topright_top, anchor_topright_right, anchor_bottomright_right, anchor_bottomright_bottom, anchor_bottomleft_bottom, anchor_bottomleft_left, anchor_topleft_left};
    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices);
    glDrawArrays(GL_LINE_LOOP, 0, 8);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::DrawPoints(const Color4& color, const Vector2* points, int num_points, float radius) {
    glPointSize(radius);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), points);
    glDrawArrays(GL_POINTS, 0, num_points);
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FIllTriangle(const Color4& color, const Vector2& triA, const Vector2& triB, const Vector2& triC) {
    FillTriangle(color, {triA, triB, triC});
}

void JGL::J2D::FillGradientTriangle(const Color4& a_color, const Color4& b_color, const Color4& c_color, const Vector2& tri_a,
                               const Vector2& tri_b, const Vector2& tri_c) {
    FillGradientTriangle(a_color, b_color, c_color, {tri_a, tri_b, tri_c});
}

void JGL::J2D::OutlineEllipse(const Color4& color, const Vector2& position, float radius_x, float radius_y, float thickness,
                         int subdivisions) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    float step = (2.f * Math::Pi) / (float) subdivisions;
    std::vector<Vector2> vertices(subdivisions);
    GLfloat angle, x, y;

    int i = 0;
    for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
        x = radius_x * std::sin(angle) + position.x;
        y = radius_y * std::cos(angle) + position.y;
        if (i < subdivisions)
            vertices[i] = {x, y};
        else
            vertices.emplace_back(x, y);
        i++;
    }

    glLineWidth(thickness);
    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glDrawArrays(GL_LINE_LOOP, 0, (int) vertices.size());
    glColor4fv(OpenGLState::baseColor);
}

void JGL::J2D::FillEllipse(const Color4& color, const Vector2& position, float radius_x, float radius_y, int subdivisions) {
    if (!inJ2D)
        Logger::Error("Drawing J2D element before J2D begin.");

    GLfloat angle, x, y;
    float step =  (2.f * Math::Pi) / (float) subdivisions;
    std::vector<Vector2> vertices(subdivisions);

    /* Most of the time the vector "vertices" is either the same size or size + 1 of the number of subdivisions.
     * Because a float is a decimal, It'd take too long to get rid of the emplace-back making us
     * wait around for the container to resize. This gets rid of it for what we can guarantee. */
    int i = 0;
    for (angle = 0.0f; angle < (2.f * Math::Pi); angle += step) {
        x = radius_x * std::sin(angle) + position.x;
        y = radius_y * std::cos(angle) + position.y;
        if (i < subdivisions)
            vertices[i] = {x, y};
        else
            vertices.emplace_back(x, y);
        i++;
    }

    glColor4ubv(color.ptr());
    glVertexPointer(2, GL_FLOAT, sizeof(Vector2), vertices.data());
    glDrawArrays(GL_TRIANGLE_FAN, 0, (int) vertices.size());
    glColor4fv(OpenGLState::baseColor);
}